// TypeScript is not capable of generating the '.js' suffix while browsers impose it (https://github.com/Microsoft/TypeScript/issues/27287):
import {Google_Images} from './Google_Images.js'; // No need of 'use strict' when 'import'
import {Open_Food_Facts} from './Open_Food_Facts.js'; // No need of 'use strict' when 'import'
import {Open_Food_Facts_item} from "./Open_Food_Facts.js";
import {Third_dimension} from './Third_dimension.js';

// Java data message:
interface Candidate_item { // Id.: 'gtin' && 'word'
    gtin: string;
    strong: number;
    weak: number;
    word: string;
    product_name?: string;
    image_URL?: string;
}

interface Displayed_image {
    index: number; // Index of candidate item in 'Carrefour_Items._All'
    image_URL?: string;
    time_out?: number;
}

// https://mariusschulz.com/blog/typing-functions-in-typescript
type Release = (n: number) => void;

class Carrefour_Items {

    private static readonly _All: Array<Candidate_item> = new Array();
    private static readonly _Number_of_Carrefour_Items_columns: number = 4;// 16;
    private static readonly _Number_of_Carrefour_Items_rows: number = 3;// 12;
    private static readonly _Carrefour_Items_images: Array<Displayed_image> = new Array(Carrefour_Items._Number_of_Carrefour_Items_columns * Carrefour_Items._Number_of_Carrefour_Items_rows).fill(-1);

    static Update(ci: Candidate_item): void {
        let index: number = Carrefour_Items._All.findIndex((element: Candidate_item) => {
            return element.gtin === ci.gtin && element.word === ci.word; // Id.: 'gtin' && 'word'
        });
        if (index === -1) { // Not found...
            index = Carrefour_Items._All.push(ci) - 1;
            // console.log("Added to 'Carrefour_Items._All': " + ci.gtin + ' at: ' + index);
            Open_Food_Facts.Get_instance(ci.gtin).get_Open_Food_Facts_item().then((offi: Open_Food_Facts_item) => {

// Look at scrapped data:
//                 window.alert(JSON.stringify(offi));

                Third_dimension.Visualize(offi,
                    offi.images.map((image, image_index: number) => { // Get a released index for each image...
                        let index_ = Carrefour_Items._Carrefour_Items_images.findIndex((element: Displayed_image) => {
                            return element.index === -1; // Released index...
                        });
                        if (index_ === -1) { // Not enough room on screen...
                            // console.log("No room for gtin: " + ci.gtin);
                            return new Promise((release: Release) => { // Index *WILL BE* booked...
                                let callback = (event: CustomEvent) => {
                                    event.stopImmediatePropagation(); // Other subscribers cannot use this index...
                                    window.removeEventListener('Release', callback); // As winning subscriber, I'm no longer interested in booking an index...

                                    // console.log("Released position " + event.detail.index + " in 'Carrefour_Items._Carrefour_Items_images' for: " + ci.gtin);

                                    Carrefour_Items._Carrefour_Items_images[event.detail.index] = {
                                        index: index, // Points to 'ci' in 'Carrefour_Items._All'...
                                        image_URL: offi.image_URLs[image_index],
                                        time_out: undefined
                                    };
                                    release(event.detail.index); // Index *IS JUST* booked, ready for visualization...
                                };
                                window.addEventListener('Release', callback); // Create subscriber...
                            });
                        } else { // Enough room on screen...
                            Carrefour_Items._Carrefour_Items_images[index_] = {
                                index: index, // Points to 'ci' in 'Carrefour_Items._All'...
                                image_URL: offi.image_URLs[image_index],
                                time_out: undefined
                            };
                            return Promise.resolve(index_); // Index *IS JUST* booked, ready for visualization...
                        }
                    })
                );
            });
        } else {
            Carrefour_Items._All[index].strong = ci.strong > Carrefour_Items._All[index].strong ? ci.strong : Carrefour_Items._All[index].strong;
            Carrefour_Items._All[index].weak = ci.weak > Carrefour_Items._All[index].weak ? ci.weak : Carrefour_Items._All[index].weak;
        }
    }

    private static readonly _Buffer: Array<string> = new Array();
    private static _Query: HTMLElement | null = null;
    // https://stackoverflow.com/questions/388996/regex-for-javascript-to-allow-only-alphanumeric
    private static readonly _Regular_expression: RegExp = /^[a-z0-9 ]+$/i; // 'Space' included...

    private static readonly _Handshaking: object = {Carrefour_Items: "Handshaking"};
    private static readonly _Raw_query: object = {Carrefour_Items: "Raw_query"};
    private static _WebSocket: WebSocket | null = null;

    static Initialization(): void {
        // Absent dans OpenFoodFacts *ET* Carrefour.fr (recherche "Cassoulet" chez Carrefour Items) :
        /*
        {
            "gtin": "3245390183958",
            "brand": "GRAND JURY",
            "capacity_factor": 1,
            "capacity_unit": "KG",
            "capacity_volume": 0.84,
            "category_structures": {
                "hyper": {
                    "hypUbDesc": "CASSOULET SUPERIEUR BOITE 4/4",
                    "hypClassDesc": "PLATS CUISINES",
                    "hypUbKey": "1466004",
                    "hypGrpClassKey": "146",
                    "hypSectorDesc": "PGC",
                    "hypDepartmentDesc": "EPICERIE",
                    "hypSectorKey": "1",
                    "hypSubClassDesc": "CASSOULET ET DERIVES",
                    "hypGrpClassDesc": "CONSERVES ET PLATS CUISINES",
                    "hypDepartmentKey": "14",
                    "hypClassKey": "1466",
                    "hypSubClassKey": "14660"
                },
                "proxy": {
                    "prxShoDepartmentDesc": "EPICERIE",
                    "prxShoDepartmentKey": "01",
                    "prxSubClassDesc": "CASSOULETS",
                    "prxShoSectorKey": "01",
                    "prxGrpClassKey": "032",
                    "prxHahSectorDesc": "EPICERIE",
                    "prxClassKey": "237",
                    "prxHahDepartmentKey": "01",
                    "prxUbDesc": "CASSOULET SUPERIEUR BOITE 4/4",
                    "prxSubClassKey": "237001",
                    "prxGrpClassDesc": "PLATS CUISINES",
                    "prxClassDesc": "PLATS CUISINES",
                    "prxUbKey": "23700134",
                    "prxHahDepartmentDesc": "EPICERIE",
                    "prxShoSectorDesc": "P. G. C.",
                    "prxHahSectorKey": "01"
                },
                "super": {
                    "supDepartmentKey": "01",
                    "supSubClassKey": "145001",
                    "supSectorKey": "1",
                    "supClassKey": "145",
                    "supSectorDesc": "DENREES NON PERISSABLES",
                    "supGrpClassDesc": "PLATS CUISINES",
                    "supClassDesc": "PLATS CUISINES",
                    "supSubClassDesc": "PLATS CUISINES FRANCAIS",
                    "supUbKey": "14500101",
                    "supDepartmentDesc": "EPICERIE",
                    "supGrpClassKey": "145",
                    "supUbDesc": "PLATS CUIS.CASSOULETS SUPERIEU"
                }
            },
            "description": "4/4 CASSOULET C.CANARD GDJ",
            "name": "4/4 CASSOULET CND"
         }
         */

        Google_Images.Get_instance('CASSOULET SUPERIEUR'); // Les produits OpenFoodFacts sont ignorés...

        // Open_Food_Facts.Get_instance("3560070478781"); // Pain d'épices Carrefour bio. -> found!

        /**
         * HTML stuff...
         */
        console.assert(window.document.readyState === 'complete'); // Anyway, 'interactive' should be sufficient (and thus better)...
        /** For test only (to be removed later on in 'index.html'): */
        Carrefour_Items._Query = window.document.getElementById('Query');

        Carrefour_Items._Carrefour_Items_images.fill({index: -1, image_URL: undefined, time_out: undefined});

        Third_dimension.Initialization(Carrefour_Items._Number_of_Carrefour_Items_columns, Carrefour_Items._Number_of_Carrefour_Items_rows);
    }

    static Connect(): void {
        // Homemade WebSockets Java library without the possibility of changing parameters' value:
        Carrefour_Items._WebSocket = new WebSocket("ws:127.0.0.1:1963", "FranckBarbier");
        Carrefour_Items._WebSocket.onmessage = (event: MessageEvent) => {
            // console.log("Message from Java: " + event.data);
            Carrefour_Items.Update(JSON.parse(event.data));
// Raw message from Java:
// '|| null' => if the first operand is truthy, it evaluates to the first operand. Otherwise, it evaluates to the second.
        };
        Carrefour_Items._WebSocket.onopen = () => {
            if (Carrefour_Items._WebSocket !== null)
                Carrefour_Items._WebSocket.send(JSON.stringify(Carrefour_Items._Handshaking));
        };
        Carrefour_Items._WebSocket.onclose = () => {
            // alert("Bye...");
        };
        Carrefour_Items._WebSocket.onerror = () => {
            // alert("Error...");
        };
    }

    static Initiate(event: CustomEvent): void {
        // console.log('Initiate: ' + JSON.stringify(event.detail) + ' gtin: ' + Carrefour_Items._All[Carrefour_Items._Carrefour_Items_images[event.detail.index].index].gtin);
        // Record time-out for possible late cancellation (see 'Validate'):
        Carrefour_Items._Carrefour_Items_images[event.detail.index].time_out = event.detail.time_out;
        Carrefour_Items._All[Carrefour_Items._Carrefour_Items_images[event.detail.index].index].product_name = event.detail.product_name;
    }

    static Invalidate(event: CustomEvent): void {
        // console.log('Invalidate: ' + JSON.stringify(event.detail));
        Carrefour_Items._Carrefour_Items_images[event.detail.index].index = -1;
        window.dispatchEvent(new CustomEvent('Release', {detail: {index: event.detail.index}}));
    }

    static Validate(event: CustomEvent): void {
        console.assert(Carrefour_Items._Carrefour_Items_images[event.detail.index].index !== -1, "'Carrefour_Items._Carrefour_Items_images[event.detail.index].index === -1', why?"); // Validated image is surely a product...
        // console.log('Validate: ' + JSON.stringify(event.detail));
        if (Carrefour_Items._Carrefour_Items_images[event.detail.index].time_out !== undefined)
            window.clearTimeout(Carrefour_Items._Carrefour_Items_images[event.detail.index].time_out);

// Tell Java to download image from image URL because of CORS in the browser that lets us get stuck:
        if (Carrefour_Items._Carrefour_Items_images[event.detail.index].image_URL !== undefined)
            Carrefour_Items._All[Carrefour_Items._Carrefour_Items_images[event.detail.index].index].image_URL = Carrefour_Items._Carrefour_Items_images[event.detail.index].image_URL;
        else
            Carrefour_Items._All[Carrefour_Items._Carrefour_Items_images[event.detail.index].index].image_URL = "";

        if (Carrefour_Items._WebSocket !== null)
            Carrefour_Items._WebSocket.send(JSON.stringify(Carrefour_Items._All[Carrefour_Items._Carrefour_Items_images[event.detail.index].index]));
    }

    static Handle(event: KeyboardEvent): void {
        if (event.defaultPrevented) { // 'event' has been already processed with stopping request
            return;
        }
        /**
         * Test
         */
        // console.log(`${event.key}` + ' <-event.key - event.code-> ' + `${event.code}`);
        // console.log('match: ' + event.key.match(Carrefour_Items._Regular_expression));
        if (Carrefour_Items._Query !== null)
            Carrefour_Items._Query.innerHTML += `${event.code}`;
        /**
         * End of test
         */
        if (event.code === 'Backspace')
            Carrefour_Items._Buffer.pop();
        if (event.code === 'Escape' || event.code === 'Enter') {
            Carrefour_Items._Buffer.length = 0;
            if (Carrefour_Items._WebSocket !== null)
                Carrefour_Items._WebSocket.send(JSON.stringify(Carrefour_Items._Raw_query));
        }
        if (Carrefour_Items._Regular_expression.test(event.key) && event.key.length === 1) {
            let last: string | undefined = Carrefour_Items._Buffer.pop();
            if (event.code === 'Space') {
                if (last !== undefined) {
                    if (!last.includes(" ")) // 'includes' tests equality as well...
                        Carrefour_Items._Buffer.push(last); // Re-inject 'last' at last position
                    Carrefour_Items._Buffer.push(event.key);
                }
                // else {
                //     console.assert(last === undefined);
                // }
            } else { // Something else that 'Space'...
                if (last === undefined)
                    Carrefour_Items._Buffer.push(event.key);
                else {
                    Carrefour_Items._Buffer.push(last); // Re-inject 'last' at last position
                    Carrefour_Items._Buffer.push(event.key);
                    if (!last.includes(" ")) { // 'includes' tests equality as well...
                        // console.log("Attempt to send to Java: " + Carrefour_Items._Buffer.join(''));
                        if (Carrefour_Items._WebSocket !== null)
                            Carrefour_Items._WebSocket.send(JSON.stringify({Carrefour_Items: Carrefour_Items._Buffer.join('')}));
                    }
                }
            }
        }
        Third_dimension.Handle(Carrefour_Items._Buffer);
    }
}

window.document.onreadystatechange = () => {
    if (window.document.readyState === 'complete') {
        Carrefour_Items.Initialization();
    }
};
window.addEventListener('load', Carrefour_Items.Connect);
window.addEventListener('keyup', Carrefour_Items.Handle); // https://developer.mozilla.org/en-US/docs/Web/Events/keyup

window.addEventListener('Initiate', Carrefour_Items.Initiate);
window.addEventListener('Invalidate', Carrefour_Items.Invalidate);
window.addEventListener('Validate', Carrefour_Items.Validate);

