// import {extract} from "./Utilities.js";

interface Google_Images_item {
    images: Array<Promise<any>>;
    image_URLs: Array<string>;
    product_name: string;
}

export class Google_Images {

    private static readonly _All: Array<Google_Images> = new Array();

    public static Get_instance(query: string): Google_Images {
        let gi: Google_Images | undefined = Google_Images._All.find((element: Google_Images) => {
            return element._query === query;
        });
        if (gi === undefined) {
            gi = new Google_Images(query);
            Google_Images._All.push(gi);
        }
        return gi;
    }

// dia, picard, système U,

    private static readonly _RH_tag: RegExp = /"rh":"(amazon\.fr|auchandirect\.fr|bienmanger\.com|carrefour\.fr|catalogue\.cerclevert\.fr|chronodrive\.com|cdiscount\.com|danival\.fr|e-leclerc\.com|episaveurs\.fr|etiketbio\.eu|geantcasino\.fr|intermarche\.com|lagrandeepicerie\.com|lesrobinsdesbio\.com|leaderprice\.fr|lidl\.fr|mangerbiochezmoi\.com|monoprix\.fr|promobutler\.be|provincesbio\.com|prixing\.fr|shoptimise\.fr|webecologie\.com)"/gi;
    private static readonly _PT_tag: RegExp = /"pt":"[^"]*/g;
    private static readonly _S_tag: RegExp = /"s":"[^"]*/g;

    private static readonly _Image_URL_pattern: RegExp = /"(http(s?):)([/|.|\w|\s|-])*\.(?:gif|jpeg|jpg|png)("|\?)/gi; // 'g' means all occurrences while 'i' means case-insensitive...
    private static readonly _Path: string = '/search?tbm=isch&q=';
    // private static readonly _Pattern: RegExp = /https:\/\/encrypted\-tbn0\.gstatic\.com/g;
    private static readonly _URL: string = 'https://cors-anywhere.herokuapp.com/https://www.google.com';

    private readonly _item_data: Promise<Google_Images_item>;

    private constructor(private readonly _query: string) { // '_query' *field* automatic inclusion
        // => this._query = _query; -> no need!
        this._item_data = new Promise((ready, problem) => { // 'problem' aims at being called when 'Promise' object has difficulty in achieving its job...
            this._get_Google_Images(ready, problem);
        });
    };

    // An 'async' function always returns a 'Promise' object (https://blog.bitsrc.io/keep-your-promises-in-typescript-using-async-await-7bdc57041308).
    // The 'Promise' object resolves by construction to the returned value:
    public async get_Google_Images() { // So returned type is inferred from 'async'...
        return await this._item_data; // Wait for '_get_item_data' (i.e., 'ready(offi);')
    }

    private _get_Google_Images(ready, problem, time_out?: number): void { // Call may omit 'time_out' optional parameter
        // https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API/Using_Fetch
        fetch(`${Google_Images._URL}${Google_Images._Path}${this._query}`).then(function (response: Response) {
            response.text().then(function (html: string) {
                    console.log('HTML: ' + html);

                    console.log(html.match(Google_Images._S_tag));

                    // Split on 'this._query': "CASSOULET GRAND JURY SUPERIEUR BOITE 4/4"
                    html.split(this._query).forEach((html_piece: string) => {
                        // if (Google_Images._RH_tag.test(html_piece)) { // Filtre sur URL des sites...
                            const image_URLs: RegExpMatchArray | null = html_piece.match(Google_Images._Image_URL_pattern);
                            if (image_URLs !== null) {
                                image_URLs.forEach((image_URL: string) => {
                                    image_URL = image_URL.substring(1, image_URL.length - 1);
                                    console.log('image_URL: ' + image_URL);
                                });
                            }
                        // }
                    });
                }.bind(this)
            )
                .catch((error: Error) => {
                    problem();
                    window.alert('Error response.json().then: ' + error.message);
                });
        }.bind(this))
            .catch((error: Error) => {
                //problem();
                window.alert('Error ' + `${Google_Images._URL}${Google_Images._Path}${this._query}: ` + error.name);
            });
    }
}
