var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { extract } from "./Utilities.js";
export var Access;
(function (Access) {
    Access[Access["Failure"] = 0] = "Failure";
    Access[Access["Success"] = 1] = "Success";
})(Access || (Access = {}));
export class Open_Food_Facts {
    constructor(_gtin) {
        this._gtin = _gtin;
        // => this._gtin = _gtin; -> no need!
        this._item_data = new Promise /*<Open_Food_Facts_item>*/((ready, problem) => {
            this._get_Open_Food_Facts_item(ready, problem);
        });
    }
    static Get_instance(gtin) {
        let off = Open_Food_Facts._All.find(element => {
            return element._gtin === gtin;
        });
        if (off === undefined) {
            // console.log("Created in 'Open_Food_Facts._All': " + gtin);
            off = new Open_Food_Facts(gtin);
            Open_Food_Facts._All.push(off);
        }
        return off;
    }
    ;
    // An 'async' function always returns a 'Promise' object (https://blog.bitsrc.io/keep-your-promises-in-typescript-using-async-await-7bdc57041308).
    // The 'Promise' object resolves by construction to the returned value:
    get_Open_Food_Facts_item() {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._item_data; // Wait for '_get_item_data' (i.e., 'ready(offi);')
        });
    }
    _get_Open_Food_Facts_item(ready, problem, time_out) {
        // https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API/Using_Fetch
        // https://en.wiki.openfoodfacts.org/API/Read/Product#Structure_of_the_call (e.g., https://ssl-api.openfoodfacts.org/api/v0/product/3664346300155.json):
        fetch(`${Open_Food_Facts._URL}${Open_Food_Facts._Path}${this._gtin}.json`, Open_Food_Facts._RequestInit)
            .then(function (response) {
            response.json()
                .then(function (result) {
                console.assert(typeof result === 'object');
                let found = result[Open_Food_Facts._Status];
                console.log(this._gtin + ' status: ' + result[Open_Food_Facts._Status]);
                if (found !== undefined && found === 1) {
                    let offi = {};
                    offi['product_name'] = extract(result, 'product', 'product_name');
                    offi['image_URLs'] = [
                        extract(result, 'product', 'image_url'),
                        extract(result, 'product', 'image_front_url'),
                        extract(result, 'product', 'image_small_url')
                    ];
                    // Remove 'null' and duplicates:
                    offi['image_URLs'] = Array.from(new Set(offi['image_URLs'].filter(image_URL => image_URL !== null)));
                    offi['images'] = [];
                    // Get images from URLs:
                    offi['image_URLs'].forEach(image_URL => {
                        console.log('Called URL: ' + image_URL);
                        // https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image
                        let candidate_image = new Image;
                        // candidate_image.crossOrigin = "anonymous"; // Default
                        offi['images'].push(new Promise((send) => {
                            candidate_image.onload = () => {
                                // console.assert(candidate_image.complete);
                                send({ access: Access.Success, image_URL: candidate_image.src });
                            };
                            candidate_image.onerror = () => {
                                send({ access: Access.Failure, image_URL: candidate_image.src });
                            };
                            candidate_image.src = image_URL;
                        }));
                        //     /**
                        //      * Fetch image is not permitted by CORS
                        //      */
                        //     let headers: Headers = new Headers();
                        //     headers.append('accept', 'image/*');
                        //     let options: RequestInit = {
                        //         headers: headers,
                        //         method: 'get'
                        //     };
                        //     fetch(image_URL, options)
                        //         .then(function (response: Response) {
                        //             response.blob()
                        //                 .then(function (blob: Blob) {
                        //                     // Process 'blob'
                        //                 }.bind(this))
                        //                 .catch(function (error: Error) {
                        //                     // This is where you run code if the server returns any errors
                        //                 });
                        //         }.bind(this))
                        //         .catch((error: Error) => {
                        //             window.alert('CORS error ' + image_URL + ': ' + error.message);
                        //         });
                        //     /**
                        //      * End of fetch image
                        //      */
                        //
                    }); // End of 'forEach(image_URL => {'
                    ready(offi);
                }
                else { // Not found
                    let offi = {};
                    offi['product_name'] = this._gtin;
                    offi['image_URLs'] = [];
                    offi['images'] = [];
                    // ready(offi);
                }
            }.bind(this))
                .catch(function (error) {
                problem();
                window.alert('Error response.json().then: ' + error.message);
            });
        }.bind(this))
            .catch(function (error) {
            problem();
            window.alert('Error ' + `${Open_Food_Facts._URL}: ` + error.message);
        });
    }
}
Open_Food_Facts._All = new Array();
Open_Food_Facts._URL = 'https://ssl-api.openfoodfacts.org';
Open_Food_Facts._Path = '/api/v0/product/';
Open_Food_Facts._RequestInit = {
    body: null,
    headers: {
        'accept': 'application/json',
        // The only allowed values for the 'Content-Type' header are 'application/x-www-form-urlencoded', 'multipart/form-data' and 'text/plain'
        // https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS
        // https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS#Simple_requests
        // "content-type": "application/json; charset=UTF-8" -> CORS problem
        'content-type': 'text/plain' // No need, anyway
    },
    method: 'get'
};
Open_Food_Facts._Status = 'status'; // '0' not found
