import {Access, Image_access, Open_Food_Facts_item} from "./Open_Food_Facts.js";

// TypeScript raises errors because 'THREE' and its content isn't known
// Either, 'THREE' and inner types are installed -> https://www.npmjs.com/package/@types/three
// or (http://blogs.microsoft.co.il/gilf/2013/07/22/quick-tip-typescript-declare-keyword/):
declare let THREE; // Type is 'any'
console.assert(typeof THREE !== "undefined", "Problem: it won't work because 'three.js' has not been loaded...");

declare let $;
console.assert(typeof $ !== "undefined", "Problem: it won't work because 'jquery-3.4.1.min.js' has not been loaded...");

export class Third_dimension {

    private static readonly _Ambient_light = new THREE.AmbientLight(0x404040);
    private static readonly _Camera = new THREE.PerspectiveCamera(50, window.innerWidth / window.innerHeight, 0.1, 20000);
    private static readonly _Renderer = new THREE.WebGLRenderer(/*{alpha: true}*/);
    private static readonly _Scene = new THREE.Scene();

    private static _Text_canvas: HTMLCanvasElement;
    private static _Text_canvas_context: CanvasRenderingContext2D | null;
    private static _Text_canvas_texture: any;

    private static _Item_geometry = new THREE.SphereBufferGeometry(10, 50, 50, 0, Math.PI);

    static Visualize(offi: Open_Food_Facts_item, indices: Array<Promise<number>>): void {
        console.assert(offi.images.length === indices.length);

        let canvas: HTMLCanvasElement = window.document.createElement('canvas');
        canvas.width = 80;
        canvas.height = 60;
        let canvas_context: CanvasRenderingContext2D | null = canvas.getContext('2d');
        if (canvas_context !== null) {
            canvas_context.font = "10px Comic Sans MS";
            canvas_context.fillStyle = "green";
            canvas_context.fillText(offi.product_name, canvas.width / 2, canvas.height / 2);
            canvas_context.textAlign = "center";
        }

        offi.images.forEach((element: Promise<Image_access>, index: number) => {
            Promise.all([element, indices[index]]).then((values: Array<any>) => {
                let image = <HTMLImageElement>window.document.getElementById("Carrefour_Items_image" + values[1]);
                image.onload = () => {

                    // https://onextrapixel.com/javascript-image-magnification-plugins/
                    // https://github.com/paulkr/blowup.js
                    $('#' + image.id).blowup({
                        "background": "#F39C12",
                        "width": 250,
                        "height": 250
                    });

                    image.style.opacity = "1";
                    let time_out: number = window.setTimeout(() => { // Image aims at disappearing...
                        image.style.opacity = "0.1";
                        window.dispatchEvent(new CustomEvent('Invalidate', {detail: {index: values[1]}}));
                    }, 10000); // 1 min.
                    window.dispatchEvent(new CustomEvent('Initiate', {
                        detail: {
                            index: values[1],
                            product_name: offi.product_name,
                            time_out: time_out
                        }
                    }));
                };
                image.src = (values[0].access === Access.Success) ? values[0].image_URL : (canvas !== null) ? canvas.toDataURL('image/png') : ''; // Load image on screen...
            });
        });
    }

    // Pas de texture utilisable à partir des images (erreurs WebGL):
    // let i = 10;
    // item.images.forEach((promise, index) => {
    //     let item_name: string = item.product_name + item.image_URLs[index];
    //     if (Third_dimension._Scene.getObjectByName(item_name) === undefined) {
    //         let group = new THREE.Group();
    //         group.name = item_name;
    //         //  Third_dimension._Scene.add(group);
    //         promise.then(image => {
    //             let back = new THREE.Mesh(Third_dimension._Item_geometry, new THREE.MeshBasicMaterial({
    //                 color: new THREE.Color('blue'),
    //
    //                 side: THREE.DoubleSide,
    //                 wireframe: true
    //             }));
    //             back.rotateY(Math.PI);
    //
    //             new Promise(ready => {
    //                 (new THREE.TextureLoader()).load(item.image_URLs[index], texture => ready(texture))
    //             }).then(texture => {
    //                 back.material.texture = texture;
    //             });
    //
    //             back.name = item_name + "back";
    //             i += 10;
    //             back.translateY(i); // Down
    //             // group.add(back);
    //             Third_dimension._Scene.add(back);
    //         });
    //     }
    // });

    static Animate(): void {
        window.requestAnimationFrame(Third_dimension.Animate);

        Third_dimension._Scene.getObjectByName('Carrefour Items').rotation.x += 0.05;
        Third_dimension._Scene.getObjectByName('Carrefour Items').rotation.y += 0.05;
        Third_dimension._Scene.getObjectByName('Carrefour Items').rotation.z += 0.05;
        Third_dimension._Scene.getObjectByName('Query').rotation.y += 0.01;

        Third_dimension._Renderer.render(Third_dimension._Scene, Third_dimension._Camera);
    }

    static Handle(buffer: Array<string>): void {
        // console.log(buffer.join('') + ' with length: ' + buffer.length);
        if (Third_dimension._Text_canvas_context !== null) {
            Third_dimension._Text_canvas_context.clearRect(0, 0, Third_dimension._Text_canvas.width, Third_dimension._Text_canvas.height);
            Third_dimension._Text_canvas_context.fillText(buffer.join(''), Third_dimension._Text_canvas.width / 2, Third_dimension._Text_canvas.height / 2);
            Third_dimension._Text_canvas_texture.needsUpdate = true;
        }
    }

    static Initialization(Number_of_Carrefour_Items_columns: number, Number_of_Carrefour_Items_rows: number): void {
        // Conversion by means of '<HTMLElement>' tells the TypeScript compiler that returned value is "obviously" not 'null'
        let images: HTMLElement = <HTMLElement>window.document.getElementById('Carrefour_Items_images');
        console.assert(images !== null, "'images === null', why?");

        images.style.display = "grid";
        images.style['grid-template-columns'] = "repeat(" + Number_of_Carrefour_Items_columns + ", 1fr)";
        images.style['grid-template-rows'] = "fit-content(" + (100 / Number_of_Carrefour_Items_rows) + "%)";
        images.style['align-items'] = "center"; /* Value of 'align-self' for children */
        images.style['justify-items'] = "flex-start"; /* Value of 'justify-self' for children */

        for (let i: number = 0; i < Number_of_Carrefour_Items_columns * Number_of_Carrefour_Items_rows; i++) {
            let image = new Image;
            image.setAttribute("id", "Carrefour_Items_image" + i);
            image.addEventListener('click', (event: Event) => {
                window.dispatchEvent(new CustomEvent('Validate', {detail: {index: i}}));
            });
            image.style.height = (100 / (Number_of_Carrefour_Items_rows)) + "vh";
            image.style.opacity = "0.5";
            image.style.width = (100 / Number_of_Carrefour_Items_columns) + "vw";

            image.onload = () => {
                images.appendChild(image); // Images are stored in a random way...
            };
            image.src = './img/Franck.jpg';
        }

        Third_dimension._Item_geometry.computeBoundingSphere();
        Third_dimension._Camera.position.z = 100;
        Third_dimension._Renderer.setClearColor(new THREE.Color('silver'), 0.1);
        Third_dimension._Renderer.setPixelRatio((window.devicePixelRatio) ? window.devicePixelRatio : 1);
// Third_dimension._Renderer.setSize(window.innerWidth, window.innerHeight);

// Casting bypasses '"strictNullChecks": true':
        let third_dimension: HTMLElement = <HTMLElement>document.getElementById('Third_dimension');

        Third_dimension._Renderer.setSize(third_dimension.clientWidth, third_dimension.clientHeight);
// http://stackoverflow.com/questions/10425310/three-js-full-screen-issue
        Third_dimension._Renderer.domElement.style.display = "block";
        third_dimension.appendChild(Third_dimension._Renderer.domElement); // Canvas généré par le 'renderer' est ajouté au DOM

        Third_dimension._Scene.add(Third_dimension._Ambient_light);

// https://gist.github.com/studioijeoma/1186920/5f50d999aa07491e67ec60d05e892637185cd4cd
        Third_dimension._Text_canvas = window.document.createElement('canvas');
        Third_dimension._Text_canvas.width = 400;
        Third_dimension._Text_canvas.height = 300;
        Third_dimension._Text_canvas_context = Third_dimension._Text_canvas.getContext('2d');

        if (Third_dimension._Text_canvas_context !== null) {
            Third_dimension._Text_canvas_context.font = "Bold 40px Arial";

            //  Third_dimension._Text_canvas_context.fillStyle = "rgba(255,0,222,0.5)";
            Third_dimension._Text_canvas_context.fillStyle = "silver";
            Third_dimension._Text_canvas_context.textAlign = "center";
            Third_dimension._Text_canvas_context.textBaseline = "middle";
            Third_dimension._Text_canvas_context.fillText('?', Third_dimension._Text_canvas.width / 2, Third_dimension._Text_canvas.height / 2);

            Third_dimension._Text_canvas_texture = new THREE.CanvasTexture(Third_dimension._Text_canvas);
            // Third_dimension._Text_canvas_texture.needsUpdate = true; // Automatic with 'THREE.CanvasTexture'
            // Third_dimension._Text_canvas_texture.minFilter = THREE.LinearFilter;
            let query_text = new THREE.Mesh(new THREE.PlaneGeometry(10, 10), new THREE.MeshBasicMaterial({
                    color: new THREE.Color(0xFFFF00),
                    map: Third_dimension._Text_canvas_texture,
                    side: THREE.DoubleSide
                }
            ));
            query_text.name = 'Query';
            query_text.translateY(-10); // Down
            query_text.quaternion.setFromAxisAngle(new THREE.Vector3(1, 0, 0), -Math.PI / 10); // Flat
            Third_dimension._Scene.add(query_text);
        }

        Third_dimension.Load();
    }

    static Load(): void {
        let lips_loaded = new Promise<any>(send => {
            new THREE.LegacyJSONLoader().load('./models/Mouth.json', lips_geometry => {
                send(lips_geometry);
            });
        }).then(lips_geometry => {
            lips_geometry.name = 'models/Mouth.json'; // Loading 'models/Mouth.json' with predefined features
            console.assert(lips_geometry.vertices.length === 3052, 'lips_geometry.vertices.length === 3052');
            console.assert(lips_geometry.faces.length === 5952, 'lips_geometry.faces.length === 5952');
            console.assert(lips_geometry.faceVertexUvs.length === 1, 'lips_geometry.faceVertexUvs.length === 1');
            console.assert(lips_geometry.faceVertexUvs[0].length === 0, 'lips_geometry.faceVertexUvs[0].length === 0'); // Only one (empty) UV layer at position '0'
            console.assert(lips_geometry.morphTargets.length === 0, 'lips_geometry.morphTargets.length === 0'); // No morphing yet

            /** Erasure of useless geometry (faces): */
            lips_geometry.faces.splice(5600, 5664 - 5600);
            lips_geometry.faces.splice(5440, 5472 - 5440);
            lips_geometry.faces.splice(4768, 5280 - 4768);
            lips_geometry.faces.splice(3488, 3744 - 3488);
            lips_geometry.faces.splice(2624, 2688 - 2624);
            lips_geometry.faces.splice(2464, 2496 - 2464);
            lips_geometry.faces.splice(1792, 2304 - 1792);
            lips_geometry.faces.splice(512, 768 - 512);
            console.assert(lips_geometry.faces.length === 4224, 'lips_geometry.faces.length === 4224');
            /** Erasure of useless geometry (vertices): */
            let indexes_of_removed_vertices = new Set(Array.apply(null, {length: lips_geometry.vertices.length}).map(Number.call, Number)); // 3052
            lips_geometry.faces.forEach(function (face) {
                indexes_of_removed_vertices.delete(face.a);
                indexes_of_removed_vertices.delete(face.b);
                indexes_of_removed_vertices.delete(face.c);
            });
            indexes_of_removed_vertices.forEach(function (index: number) {
                delete lips_geometry.vertices[index]; // Removed vertices are set to 'undefined' before definitive erasure
            });
            let n;
            lips_geometry.faces.forEach(function (face) {
                n = 0;
                for (let i = 0; i < face.a; i++) {
                    if (lips_geometry.vertices[i] === undefined)
                        n++;
                }
                face.a -= n;
                n = 0;
                for (let i = 0; i < face.b; i++) {
                    if (lips_geometry.vertices[i] === undefined)
                        n++;
                }
                face.b -= n;
                n = 0;
                for (let i = 0; i < face.c; i++) {
                    if (lips_geometry.vertices[i] === undefined)
                        n++;
                }
                face.c -= n;
            });
            lips_geometry.vertices = lips_geometry.vertices.filter(function (element) {
                return element !== undefined;
            });
            console.assert(lips_geometry.vertices.length === 2188, 'lips_geometry.vertices.length === 2188');

            /** TEST: check that faces point to something correct : OK */
//        lips_geometry.faces.forEach(function (face) {
//            if (lips_geometry.vertices[face.a] === undefined || lips_geometry.vertices[face.a] === null
//                    || lips_geometry.vertices[face.b] === undefined || lips_geometry.vertices[face.b] === null
//                    || lips_geometry.vertices[face.c] === undefined || lips_geometry.vertices[face.c] === null)
//                alert(face.a + " " + face.b + " " + face.c + "*********" + JSON.stringify(lips_geometry.vertices[face.a]) + " " + JSON.stringify(lips_geometry.vertices[face.b]) + " " + JSON.stringify(lips_geometry.vertices[face.c]));
//        });
            /** End of TEST */

            /** Reordering of faces: */
            lips_geometry.faces = [].concat(
                // Outer contour (magenta):
                lips_geometry.faces.slice(512, 1024),
                lips_geometry.faces.slice(1280, 1632),
                lips_geometry.faces.slice(1696, 1760),
                lips_geometry.faces.slice(1792, 1824),
                lips_geometry.faces.slice(2624, 3136),
                lips_geometry.faces.slice(3392, 3744),
                lips_geometry.faces.slice(3808, 3872),
                lips_geometry.faces.slice(3904, 3936),
                // Red contour:
                lips_geometry.faces.slice(0, 256),
                lips_geometry.faces.slice(1632, 1664),
                lips_geometry.faces.slice(2112, 2368),
                lips_geometry.faces.slice(3744, 3776),
                // Green contour:
                lips_geometry.faces.slice(1824, 2112),
                lips_geometry.faces.slice(3936, 4224),
                // Yellow contour:
                lips_geometry.faces.slice(1024, 1280),
                lips_geometry.faces.slice(1760, 1792),
                lips_geometry.faces.slice(3136, 3392),
                lips_geometry.faces.slice(3872, 3904),
                // Inner contour (blue):
                lips_geometry.faces.slice(256, 512),
                lips_geometry.faces.slice(1664, 1696),
                lips_geometry.faces.slice(2368, 2624),
                lips_geometry.faces.slice(3776, 3808)
            );
            let face_sort = function (face1, face2) {
// Faces are ordered from lips' top to lips' bottom:
                return Math.max(lips_geometry.vertices[face2.a].length(), lips_geometry.vertices[face2.b].length(), lips_geometry.vertices[face2.c].length())
                    - Math.max(lips_geometry.vertices[face1.a].length(), lips_geometry.vertices[face1.b].length(), lips_geometry.vertices[face1.c].length());
            };

// http://stackoverflow.com/questions/7426563/in-javascript-how-do-you-sort-a-subset-of-an-array
            // Outer lips' contour (magenta):
            let post_sorted = lips_geometry.faces.slice(1920);
            lips_geometry.faces.length = 1920;
            lips_geometry.faces.sort(face_sort).push.apply(lips_geometry.faces, post_sorted);
            // Red contour:
            let pre_sorted = lips_geometry.faces.slice(0, 1920);
            post_sorted = lips_geometry.faces.slice(2496);
            let sorted = lips_geometry.faces.slice(1920, 2496).sort(face_sort);
            lips_geometry.faces.length = 0;
            lips_geometry.faces.push.apply(lips_geometry.faces, pre_sorted.concat(sorted).concat(post_sorted));
            // Green contour:
            pre_sorted = lips_geometry.faces.slice(0, 2496);
            post_sorted = lips_geometry.faces.slice(3072);
            sorted = lips_geometry.faces.slice(2496, 3072).sort(face_sort);
            lips_geometry.faces.length = 0;
            lips_geometry.faces.push.apply(lips_geometry.faces, pre_sorted.concat(sorted).concat(post_sorted));
            // Yellow contour:
            pre_sorted = lips_geometry.faces.slice(0, 3072);
            post_sorted = lips_geometry.faces.slice(3648);
            sorted = lips_geometry.faces.slice(3072, 3648).sort(face_sort);
            lips_geometry.faces.length = 0;
            lips_geometry.faces.push.apply(lips_geometry.faces, pre_sorted.concat(sorted).concat(post_sorted));
            // Blue contour:
            pre_sorted = lips_geometry.faces.slice(0, 3648);
            post_sorted = lips_geometry.faces.slice(4224);
            sorted = lips_geometry.faces.slice(3648, 4224).sort(face_sort);
            lips_geometry.faces.length = 0;
            lips_geometry.faces.push.apply(lips_geometry.faces, pre_sorted.concat(sorted).concat(post_sorted));

            /** TEST: coloration: */
            for (let i = 0; i < 1920; i++) {
                lips_geometry.faces[i].color.setRGB(0, 0, 0);
            }
            for (let i = 1920; i < 2496; i++) {
                lips_geometry.faces[i].color.setRGB(1, 0, 0);
            }
            for (let i = 2496; i < 3072; i++) {
                lips_geometry.faces[i].color.setRGB(0, 1, 0);
            }
            for (let i = 3072; i < 3648; i++) {
                lips_geometry.faces[i].color.setRGB(255, 255, 0);
            }
            for (let i = 3648; i < 4224; i++) {
                lips_geometry.faces[i].color.setRGB(0, 0, 1);
            }
            /** End of TEST */

            /** TEST: progressive coloration of most outermost contour from blue to black: */
            // let min = 0;
            // let max = 1920;
            // let color_faces = function (choice) {
            //     for (let i = min; i < choice; i++) {
            //         lips_geometry.faces[i].color.setRGB(255, 255, 255); // White
            //     }
            //     for (let i = choice; i < max; i++) {
            //         lips_geometry.faces[i].color.setRGB(0, 0, 0); // Black
            //     }
            //     lips_geometry.colorsNeedUpdate = true; // BTW, the 'geometry' is shared with 'this._lips.geometry'
            // };
            // let Lips_tester = function () {
            //     this.Color = min;
            // };
            // let dat_gui = new dat.GUI({width: 100});
            // let controller = dat_gui.add(new Lips_tester(), 'Color', min, max).step(1);
            // controller.onChange(color_faces);
            /** End of TEST */

            // Morphing preparation:
            lips_geometry.lips_outside_top = Array.from(new Set(lips_geometry.faces.slice(0, 960).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));
            lips_geometry.lips_outside_bottom = Array.from(new Set(lips_geometry.faces.slice(960, 1920).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));
            lips_geometry.outermost_upper_lips = Array.from(new Set(lips_geometry.faces.slice(1920, 2208).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));
            lips_geometry.outermost_lower_lips = Array.from(new Set(lips_geometry.faces.slice(2208, 2496).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));
            lips_geometry.upper_lips = Array.from(new Set(lips_geometry.faces.slice(2496, 2784).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));
            lips_geometry.lower_lips = Array.from(new Set(lips_geometry.faces.slice(2784, 3072).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));
            lips_geometry.innermost_upper_lips = Array.from(new Set(lips_geometry.faces.slice(3072, 3360).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));
            lips_geometry.innermost_lower_lips = Array.from(new Set(lips_geometry.faces.slice(3360, 3648).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));
            lips_geometry.lips_inside = Array.from(new Set(lips_geometry.faces.slice(3648, 4224).map(function (face) {
                return [face.a, face.b, face.c];
            }).reduce(function (previous, current) {
                return previous.concat(current);
            })));

            let mesh = new THREE.Mesh(lips_geometry, new THREE.MeshBasicMaterial({
                color: new THREE.Color("rgb(255,255,255)"), // White
                side: THREE.DoubleSide,
                vertexColors: THREE.FaceColors,
                wireframe: true
            }));
            mesh.name = "Lips geometry";
            mesh.scale.set(10, 10, 10); // Requires greater size...
// Positioning:
            lips_geometry.computeBoundingBox();
            lips_geometry.center();
// Ready:
            return mesh;
        });


        let text_loaded = new Promise<unknown>(send => {
            new THREE.FontLoader().load('./fonts/helvetiker_regular.typeface.json', font => {
                send(new THREE.TextGeometry('Carrefour Items', {
                    font: font,
                    size: 5,
                    height: 2,
                    //curveSegments: 20,
                    bevelEnabled: false,
                    bevelThickness: 5,
                    bevelSize: 5,
                    bevelSegments: 5
                }));
            });
        }).then((text_geometry: any) => {
            text_geometry.name = 'Carrefour Items';

            text_geometry.computeBoundingBox();
            text_geometry.center();

            let mesh = new THREE.Mesh(text_geometry, new THREE.MeshBasicMaterial({
                color: new THREE.Color("rgb(255,0,0)"), // Red
                // side: THREE.DoubleSide,
                vertexColors: THREE.FaceColors,
                wireframe: true
            }));
            mesh.name = text_geometry.name;
            return mesh;
        });

        Promise.all([lips_loaded, text_loaded]).then(meshes => {
            meshes.forEach(mesh => {
                Third_dimension._Scene.add(mesh);
            });
            Third_dimension.Animate();
        });
    }
}
