/* Node.js program */
'use strict';

for (let i = 0; i < process.argv.length; i++) {
    console.log(i + ' -> ' + process.argv[i]);
}

const fs = require('fs');

// 'puppeteer.js' library: https://developers.google.com/web/tools/puppeteer/
const puppeteer = require('puppeteer'); // 'npm install --save puppeteer' or 'sudo npm install --save puppeteer'

// Recent download problems require this: 'sudo npm install puppeteer --unsafe-perm=true --allow-root'

// 'request.js' library: https://www.npmjs.com/package/request
const request = require('request'); // 'npm install --save request' or 'sudo npm install --save request'

// Check whether 'puppeteer.js' and 'request.js' have been installed: 'npm list -depth=0'

class Carrefour_fr {
    static Database() {
        return '../Browser/database/';
    }
    static Path() {
        return '/s?q=';
    }
    static URL() {
        return 'https://www.carrefour.fr';
    }
    static async Take_picture() { // 'async' returns an 'AsyncFunction' object
        /* Key rule:
         * 'await' returns the promise result, say what is computed by 'puppeteer.launch'
         * *NOT USING* 'await' leads to get the promise itself!
         */
//    let call_without_await = puppeteer.launch(); // Bad idea...
//    console.log(call_without_await instanceof Promise); // 'true' is displayed!
        const browser = await puppeteer.launch({/*executablePath: '/Application/Google\ Chrome.app/Contents/MacOS',*/ headless: false}); // Launch 'Chrome' *with* visual environment
        // Following call *DOES NOT* occur (because of prior 'await') until prior one terminates:
        const page = await browser.newPage();
        // 'https://www.carrefour.fr/s?q=' seule donne le nombre total d'items !
        await page.goto(Carrefour_fr.URL() + Carrefour_fr.Path() + (typeof process.argv[2] === 'string' ? process.argv[2] : '')); // By default, 'page.goto(...)' waits for the page load event
        let photo_file_name = (typeof process.argv[2] === 'string' ? process.argv[2] : 'undefined') + '.png';

        let body = await page.evaluate(() => document.body.innerHTML);
        /** Image URLs are at a given location: */
        const start = /["']images["']\s?:\s?\[\s?{/;
        const end = /]\s?}/;
        const image_URLs_location = body.substring(body.search(start), body.search(end));
//    console.log(image_URLs_location + '\n\n');

        // Raw sought string: '/media\/1500x1500\/Photosite\/PGC\/EPICERIE\/3560070478781_PHOTOSITE_20171208_164121_0.jpg?placeholder='
        // https://javascript.info/regular-expressions
//    const pattern = /\/media\\\/\d+x\d+\\\/(\b\w+\b\\\/)+/;
        // http://www.javascripter.net/faq/backslashinregularexpressions.htm
        const pattern = new RegExp('\/media\\\\/\\d+x\\d+\\\\/(\\b\\w+\\b\\\\/)+' + process.argv[2] + /* Checking enforcement -> */ '[\\w+]'); // Replace 'media' by '\\b\\w+\\b'?
//    console.log("Found? " + pattern.test(image_URLs_location)); // 'true'?
        const start_index = image_URLs_location.search(pattern);
        if (start_index !== -1) {
            const end_index = image_URLs_location.search(/\?placeholder=/);
            // First sought occurrence for simplification:
            const image_URL = Carrefour_fr.URL() + image_URLs_location.substring(start_index, end_index).replace(/\\/g, '');
//        console.log(image_URL);
            request(image_URL).pipe(fs.createWriteStream(Carrefour_fr.Database() + photo_file_name));
        } else
            await page.screenshot({path: Carrefour_fr.Database() + photo_file_name}); // Default picture...
        await browser.close();
        return photo_file_name;
    }
}

Carrefour_fr.Take_picture().then((photo_file_name) => { // Instead of 'take_picture();', which is the "common" style of calling a function...
    console.log('Photo file name: ' + photo_file_name);
});
